@@ -39,10 +39,14 @@ THE SOFTWARE.
 #include <sys/socket.h>
 #include <linux/netlink.h>
 #include <linux/rtnetlink.h>
-#include <linux/if_bridge.h>
 #include <linux/fib_rules.h>
 #include <net/if_arp.h>
 
+/* From <linux/if_bridge.h> */
+#ifndef BRCTL_GET_BRIDGES
+#define BRCTL_GET_BRIDGES 1
+#endif
+
 #if(__GLIBC__ < 2) || (__GLIBC__ == 2 && __GLIBC_MINOR__ <= 5)
 #define RTA_TABLE 15
 #endif
@@ -951,12 +955,12 @@ kernel_route(int operation, int table,
     /* Check that the protocol family is consistent. */
     if(plen >= 96 && v4mapped(dest)) {
         if(!v4mapped(gate) ||
-           (src_plen > 0 && (!v4mapped(src) || src_plen < 96))) {
+           !v4mapped(src)) {
             errno = EINVAL;
             return -1;
         }
     } else {
-        if(v4mapped(gate)|| (src_plen > 0 && v4mapped(src))) {
+        if(v4mapped(gate) || v4mapped(src)) {
             errno = EINVAL;
             return -1;
         }
@@ -991,7 +995,7 @@ kernel_route(int operation, int table,
 
 
     ipv4 = v4mapped(gate);
-    use_src = (src_plen != 0 && kernel_disambiguate(ipv4));
+    use_src = (!is_default(src, src_plen) && kernel_disambiguate(ipv4));
 
     kdebugf("kernel_route: %s %s from %s "
             "table %d metric %d dev %d nexthop %s\n",
@@ -1092,6 +1096,7 @@ parse_kernel_route_rta(struct rtmsg *rtm, int len, struct kernel_route *route)
         /* if RTA_DST is not a TLV, that's a default destination */
         const unsigned char zeroes[4] = {0, 0, 0, 0};
         v4tov6(route->prefix, zeroes);
+        v4tov6(route->src_prefix, zeroes);
         route->plen = 96;
     }
     route->proto = rtm->rtm_protocol;
@@ -1307,10 +1312,13 @@ parse_addr_rta(struct ifaddrmsg *addr, int len, struct in6_addr *res)
     struct rtattr *rta;
     len -= NLMSG_ALIGN(sizeof(*addr));
     rta = IFA_RTA(addr);
+    int has_local = 0; /* A _LOCAL TLV may be bound with a _ADDRESS' which
+        represents the peer's address.  In this case, ignore _ADDRESS. */
 
     while(RTA_OK(rta, len)) {
         switch(rta->rta_type) {
         case IFA_LOCAL:
+            has_local = 1;
         case IFA_ADDRESS:
             switch(addr->ifa_family) {
             case AF_INET:
@@ -1327,6 +1335,8 @@ parse_addr_rta(struct ifaddrmsg *addr, int len, struct in6_addr *res)
                 return -1;
                 break;
             }
+            if(has_local)
+                return 0;
             break;
         default:
             break;
